#include "zstring.h"


//----------------------------

class wiki2html
    {
    public :
    wiki2html () ;
    void setWikiStyle ( zstring ws ) ;
    void read_command_line () ;
    void read_file ( string filename ) ;
    void do_parse () ;
    zstring get_wiki () { return wiki ; }
    
    private :
    void add_html_wiki ( zstring html , zstring open , zstring close ) ;
    zstring get_default ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close ) ;
    zstring collapse_params ( vector <zstring> &p , vector <zstring> &v ) ;

    void init_wikipedia () ;

    zstring to_wiki ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close ) ;
    zstring to_wikipedia ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close ) ;

    zstring wiki , html ;
    vector <zstring> c_html , c_wiki_open , c_wiki_close ;
    zstring wikiStyle ;
    } ;
    
wiki2html::wiki2html ()
    {
    setWikiStyle ( "WIKIPEDIA" ) ; // Default
    }
    
zstring wiki2html::to_wiki ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close )
    {
    if ( wikiStyle == "WIKIPEDIA" ) return to_wikipedia ( h , p , v , close ) ;
    }
    
void wiki2html::setWikiStyle ( zstring ws )
    {
    ws.to_upper () ;
    wikiStyle = ws ;
    c_html.clear () ;
    c_wiki_open.clear () ;
    c_wiki_close.clear () ;
    if ( wikiStyle == "WIKIPEDIA" ) init_wikipedia () ;
    }

void wiki2html::read_command_line ()
    {
    string tmp ;
    html = "" ;
    while ( !cin.eof() )
       {
       getline ( cin , tmp ) ;
       html += tmp ;
       html.push_back ( '\n' ) ;
       }
    }

void wiki2html::read_file ( string filename )
    {
    ifstream in ( filename.c_str() , ios::in ) ;
    string tmp ;
    html = "" ;
    while ( !in.eof() )
       {
       getline ( in , tmp ) ;
       html += tmp ;
       html.push_back ( '\n' ) ;
       }
    }

void wiki2html::add_html_wiki ( zstring html , zstring open , zstring close )
    {
    c_html.push_back ( html ) ;
    c_wiki_open.push_back ( open ) ;
    c_wiki_close.push_back ( close ) ;
    }    
    
void wiki2html::init_wikipedia ()
    {
    // '?' will be replaced by parameters
    add_html_wiki ( "B" , "'''" , "'''" ) ;
    add_html_wiki ( "I" , "''" , "''" ) ;
    add_html_wiki ( "P" , "\n\n" , "" ) ;
    add_html_wiki ( "H1" , "== " , " ==" ) ;
    add_html_wiki ( "H2" , "=== " , " ===" ) ;
    add_html_wiki ( "H3" , "==== " , " ====" ) ;
    add_html_wiki ( "H4" , "===== " , " =====" ) ;
    add_html_wiki ( "TABLE" , "\n{| ?" , "\n|}" ) ;
    add_html_wiki ( "TD" , "\n|?|" , "" ) ;
    add_html_wiki ( "TR" , "\n|- ?" , "" ) ;
    add_html_wiki ( "TH" , "\n!?|" , "" ) ;
    add_html_wiki ( "CAPTION" , "\n|+?|" , "" ) ;
    add_html_wiki ( "BR" , "<br/>\n" , "" ) ;
    }
    
zstring wiki2html::collapse_params ( vector <zstring> &p , vector <zstring> &v )
    {
    zstring ret , dummy ;
    for ( int a = 0 ; a < p.size() ; a++ )
        {
        if ( !ret.empty() ) ret += " " ;
        dummy = p[a] ;
        dummy.to_lower() ;
        ret += dummy ;
        if ( !v[a].empty() )
           {
           dummy = v[a] ;
           dummy.to_lower() ;
           ret += "=\"" ;
           ret += dummy ;
           ret += "\"" ;
           }
        }
    return ret ;
    }
    
zstring wiki2html::get_default ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close )
    {
    for ( int a = 0 ; a < c_html.size() ; a++ )
        {
        if ( c_html[a] == h )
           {
           zstring r = close ? c_wiki_close[a] : c_wiki_open[a] ;
           zstring r1 = r.before_first ( '?' ) ;
           if ( r == r1 ) return r ;
           zstring r2 = r.after_first ( '?' ) ;
           r1 += collapse_params ( p , v ) ;
           r1 += r2 ;
           return r1 ;
           }
        }
    return "@" ; // Dummy ;
    }

zstring wiki2html::to_wikipedia ( zstring h , vector <zstring> &p , vector <zstring> &v , bool close )
    {
    zstring ret = get_default ( h , p , v , close ) ;
    if ( ret != "@" ) return ret ;
    
    ret = "<" ;
    ret += h ;
    ret += " " ;
    ret += collapse_params ( p , v ) ;
    ret[ret.length()-1] = '>' ;
    return ret ;
    }
    
void wiki2html::do_parse ()
    {
    // Preparing
    zstring s = html ;
    s.erase_all_double ( ' ' ) ;
    
    vector <zstring> valid_html ;
    valid_html.push_back ( "HTML" ) ;
    valid_html.push_back ( "HEAD" ) ;
    valid_html.push_back ( "BODY" ) ;
    valid_html.push_back ( "A" ) ;
    valid_html.push_back ( "B" ) ;
    valid_html.push_back ( "I" ) ;
    valid_html.push_back ( "P" ) ;
    valid_html.push_back ( "U" ) ;
    valid_html.push_back ( "BR" ) ;
    valid_html.push_back ( "TABLE" ) ;
    valid_html.push_back ( "CAPTION" ) ;
    valid_html.push_back ( "TR" ) ;
    valid_html.push_back ( "TH" ) ;
    valid_html.push_back ( "TD" ) ;
    valid_html.push_back ( "FONT" ) ;
    valid_html.push_back ( "SPAN" ) ;
    valid_html.push_back ( "DIV" ) ;
    valid_html.push_back ( "CENTER" ) ;
    valid_html.push_back ( "SMALL" ) ;
    valid_html.push_back ( "H1" ) ;
    valid_html.push_back ( "H2" ) ;
    valid_html.push_back ( "H3" ) ;
    valid_html.push_back ( "H4" ) ;
    valid_html.push_back ( "H5" ) ;
    valid_html.push_back ( "H6" ) ;


    int a ;
    zstring out ;
    while ( !s.empty() )
        {
        zstring q1 = s.before_first ( '<' ) ;
        zstring q2 = s.after_first ( '<' ) ;
        for ( a = 0 ; a < q2.length() && q2[a] != ' ' && q2[a] != '>' ; a++ ) ;
        zstring h = q2.left ( a ) ;
        h.to_upper() ;
        int hlen = h.length() ;
        bool close = false ;
        if ( !h.empty() && h[0] == '/' )
           {
           close = true ;
           h = h.mid ( 1 ) ;
           hlen++ ;
           }
        if ( !h.empty() && h[h.length()-1] == '/' )
           {
           h = h.left ( h.length() - 1 ) ;
//           hlen++ ;
           }
           
        bool justpaste = true ;
        int x ;
        if ( h.is_in ( valid_html ) )
           {
           out += q1 ;
           q2 = q2.mid ( hlen ) ;
           x = q2.get_first_unquoted ( '>' ) ;
           if ( x < q2.get_first_unquoted ( '<' ) )
              justpaste = false ;
           }
           
        if ( justpaste ) // No valid HTML
           {
           out += q1 ;
           if ( q2.get_first_unquoted ( '>' ) != q2.length() )
              out += "<" ;
           s = q2 ;
           }
        else
           {
           zstring params = q2.left ( x ) ;           
           s = q2.mid ( params.length()+1 ) ;
           
           vector <zstring> p , v ;
           params.erase_all_double ( ' ' , '=' ) ;
           params.erase_all_double ( '=' , ' ' ) ;
           params.trim () ;
           while ( !params.empty() )
              {
              x = params.get_first_unquoted ( ' ' ) ;
              zstring p1 = params.left ( x ) ;
              zstring p2 = p1.after_first ( '=' ) ;
              p1 = p1.before_first ( '=' ) ;
              p1.to_upper () ;
              p2.trim() ;
              p2.trim_quotes () ;
              p2.trim () ;
              p2.to_upper () ;
              params = params.mid ( x ) ;
              params.trim () ;
              p.push_back ( p1 ) ;
              v.push_back ( p2 ) ;
              }
              
           zstring result = to_wiki ( h , p , v , close ) ;
           out += result ;
           }
        }
        
    // Beautify
    out.replace ( "\n\n{|" , "\n{|" ) ;
    out.replace ( "\n\n|" , "\n|" ) ;
    out.replace ( "\n||" , "\n|" ) ;
    out.replace ( "\n!|" , "\n!" ) ;
    out.replace ( "\n|+|" , "\n|+" ) ;
    out.replace ( "<br/>\n\n" , "<br/>\n" ) ;
    
    while ( !out.empty() && out[0] == '\n' ) out = out.mid ( 1 ) ;
    while ( !out.empty() && out[out.length()-1] == '\n' ) out = out.mid ( 0 , out.length()-2 ) ;
    
    vector <zstring> v ;
    out.explode ( '\n' , v ) ;
    
    for ( a = 0 ; a < v.size() ; a++ )
        {
        v[a].trim ( false , true ) ;
        if ( a > 0 && v[a].length() > 1 && v[a-1].length() > 1 &&
                v[a-1][0] == '|' && v[a][0] == '|' && 
                v[a-1][1] != '-' && v[a][1] != '-' &&
                v[a-1][1] != '}' && v[a][1] != '}' &&
                v[a-1][1] != '+' && v[a][1] != '+' &&
                v[a-1].length() + v[a].length() < 80 )
           {
           v[a-1] += "|" ;
           v[a-1] += v[a] ;
           v.erase ( v.begin() + a ) ;
           a-- ;
           }
        else if ( a > 0 && v[a].length() > 1 && v[a-1].length() > 1 &&
                v[a-1].left(2) == "{|" && v[a] == "|-" )
           {
           v.erase ( v.begin() + a ) ;
           a-- ;
           }
                
        }
    out.implode ( '\n' , v ) ;
    while ( !out.empty() && out[out.length()-1] == '\n' ) out.pop_back() ;
    wiki = out ;
    }

int main(int argc, char *argv[])
    {
    wiki2html w2h ;
    w2h.read_command_line () ;
    w2h.do_parse () ;
    cout << w2h.get_wiki() << endl ;    
    return 0;
    }
    
